// Time-stamp: <2002-07-09 13:29:47 zaykin> [written by Dmitri Zaykin]
//
// Inferring haplotype information through the Expectation-Maximization
// algorithm and performing trait-haplotype multiple regression 
// for quantitative diploid data.
//
// Compilation:
//
// g++ -o emgi.x emgi.cpp -Wall std/std.a std/newmat10/libnewmat.a std/nr/nr.a -O3 -s
//
// std/std.a : Run Makelib.sh in ./std/ to compile this library
// std/newmat10/libnewmat.a : cd std/newmat10; make -f gnu.mak libnewmat.a
// std/nr/nr.a : Numerical Recipes in C library; from www.nr.com (need just two functions, gammln and betacf)

#include <vector>
#include <set>
#include <map>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <string>
#include <iomanip>
#include <assert.h>
#include <math.h>
#include <fstream>
#include <stdio.h>
#include "std/gauss.h"
#include "std/mystrings.h"
#include "std/mymem.h"
#include "std/attic.h"

// Numerical Recipes in C header file
#include "std/nr/nr.h"
// C++ matrix classes from newmat10 by Robert Davies
#include "std/newmat10/newmat.h"
#include "std/newmat10/newmatap.h"

typedef vector <int> Vi;
typedef vector <double> Vd;
typedef vector <Vd> Vvd;
typedef set <Vi> Svi;
typedef vector<Svi> Vsvi;
typedef map <Vi, double, less<Vi> > Md;
typedef size_t st;

Gauss Ga;
const double IgnoreFreq=0.001;

ostream& mt (ostream& os)
{
    os << setw(12) << setprecision(8);
    os << setiosflags(ios::fixed);
    return os;
}

double AdjTail(double prob, int tail=2)
{
    return prob <= 0.5 ? tail*prob : tail*(1.0-prob);
}

struct Dude // struct to hold person's genotype
{
    double rsp;  // response
    Vi g[2];
};

struct Hap
{
    Svi s; // set of all possible gametes
    Md m;  // probabilities mapped by gametes
    Vd tg; // contributions for each gamete type
};

struct Gams 
{
    double p[2];
    int h;
    Vi v, w, g[2];
    Gams() {
        p[0] = p[1] = 0;
        h = 0;
    }
    void Update(const Vi& ix) {
        v = ix;
        w.resize(v.size());
        for(st i=0; i<v.size(); i++) w[i] = !v[i];
    }
    void SetGnt(const Vi& a, const Vi& b) {
        assert(a.size() == b.size());
        assert(a.size() == v.size() && v.size() == w.size());
        g[0] = a;
        g[1] = b;
        if(g[0] > g[1]) swap(g[0], g[1]);

            // replace 0's and 1's with actual alleles
        h=0;
        for(st i=0; i<g[0].size(); i++) {
            v[i] = g[v[i]][i];
            w[i] = g[w[i]][i];
            if(v[i] != w[i]) ++h;
        }
    }
        // sort by the highest makeup probability
    int operator > (const Gams& w) const { 
        return ((p[0]+p[1]) > (w.p[0]+w.p[1]));
    }

        // "ABC" type of sort
    int operator < (const Gams& w) const { 
        return (v < w.v);
    }

    int operator == (const Gams& w) const { 
        return (v == w.v /* || v == w.w */);
    }
};

typedef vector<Gams> Vg;
typedef vector<Vg> Vvg;
typedef vector<Dude> Vdu;

// Generate Dirichlet random vector
void Diri(Vd& p, double pr)
{
    double sum=0;
    for(st j=0; j<p.size(); j++) {
        p[j] = Gengam(1.0, pr, Ga);
        sum += p[j];
    }
    for(st j=0; j<p.size(); j++) p[j] /= sum;
}

// Recursive 2^(L-1) expansion
void Enumerate(const Vi gnt[2], Vg& vg, Vi& in, int Depth, 
               int up, int curr=0)
{
    if(curr < Depth-1) { // "curr < Depth" would do all L
        for(in[curr] = 0; in[curr] < up; in[curr]++) {
            Enumerate(gnt, vg, in, Depth, up, curr+1);
        }
    }
    else {
        Gams gam;
        gam.Update(in);
        vg.push_back(gam);
    }
}

// Expand each person's genotype into all possible gametic phases
void Expand(Vg& holder, Vg& vg, const Vi gnt[2], double diri_par)
{
    const int n=gnt[0].size();
    Vi idx(n);

        // generate sets of gametic pairs
    if(holder.size() == 0) {
        Enumerate(gnt, holder, idx, n, 2);
    }

    if(vg.size() != holder.size()) vg = holder;

    for(st i=0; i<holder.size(); i++) {
        vg[i].SetGnt(gnt[0], gnt[1]);
    }
}

void GetGametes(Vg& holder, Vvg& vvg, const Vdu& vdu, double diri_par)
{
    if(vvg.size() != vdu.size()) vvg.resize(vdu.size());
    
    for(st u=0; u<vdu.size(); u++) {
        Vg vg;
        Expand(holder, vg, vdu[u].g, diri_par);
        vvg[u] = vg;
    }
}

void ReadGnt(Vdu& g, st sz, const string& miss, istream& Cin)
{
    g.clear();
    while(Cin)
    {
        Vi v[2];
        double response;
        if(!(Cin >> response)) break;

        int found_miss=0;
        for(st i=0; i < sz; i++) {
            int ii, jj;
            string iis, jjs;
            if(Cin >> iis >> jjs) {
                if(found_miss) continue;
                if(iis == miss || jjs == miss) {
                    v[0].resize(0);
                    v[1].resize(0);
                    found_miss=1;
                }
                else {
                    ii = atoi(iis.c_str());
                    jj = atoi(jjs.c_str());
                    if(ii > jj) swap(ii, jj);
                    v[0].push_back(ii);
                    v[1].push_back(jj);
                }
            }
            else break;
        }
        if(v[0].size()==sz && v[1].size()==sz) {
            Dude d;
            d.rsp = response;
            d.g[0] = v[0];
            d.g[1] = v[1];
            g.push_back(d);
        }
        else if(v[0].size() || v[1].size()) {
            cerr << "Yer matrix is ragged (exiting)\n";
            exit(1);
        }
    }
}

void Reconstruct(Vi& dst, const Vi& src, const Vi g[2])
{
    for(st i=0; i<src.size(); i++) {
        if(src[i] == g[0][i]) dst[i] = g[1][i];
        else if(src[i] == g[1][i]) dst[i] = g[0][i];
        else {   // impossible happened
            cerr << "Something is very wrong in ";
            cerr << __FILE__ << "; line " << __LINE__ << '\n';
            exit(1);
        }
    }
}

void CountEgs(Md& eg, Hap& hp, Vd& spr, const Vvg& vvg, 
              const Vsvi& vsvi)
{
    for(Svi::iterator i=hp.s.begin(); i != hp.s.end(); i++) 
    {
        double f = hp.m[*i];
        double tmp_eg = 0;
        for(st u=0; u<vvg.size(); u++) {
            if(spr[u] && vsvi[u].find(*i) != vsvi[u].end()) {
                Vi dst(i->size());
                    // every object in vvg[u] has the same g[2]
                    // so just take vvg[u][0]
                Reconstruct(dst, *i, vvg[u][0].g);
                double g = hp.m[dst];
                int nc=0;
                for(st j=0; j<vvg[u].size(); j++) {
                    if(vvg[u][j].w == dst) nc += 1;
                    if(vvg[u][j].v == dst) nc += 1;
                }
                g *= nc;
                tmp_eg += f*g / spr[u];
            }
        }
        eg[*i] = tmp_eg;
    }
}
    // start EM process
void Initiate(Md& eg, Hap& hp, Vd& spr, double& lhd0, 
              Vvg& vvg, const Vsvi& vsvi, double diri_par)
{
    hp.s.clear();
    hp.m.clear();
    
    for(st u=0; u<vvg.size(); u++) { // create set of all sample haps
        Vg &vg = vvg[u];
        for(st i=0; i<vg.size(); i++) {
            hp.s.insert(vg[i].v); 
            hp.s.insert(vg[i].w);
        }
    }
        // generate initial probabilities from a symmetric
        // Dir(diri_par) distribution; high values of diri_par result
        // in more evenly distributed probabilities

    if(hp.tg.size() != hp.s.size()) hp.tg.resize(hp.s.size());
    Diri(hp.tg, diri_par);
    
    int j=0;
    for(Svi::iterator i=hp.s.begin(); i != hp.s.end(); i++) {
        hp.m[*i] = hp.tg[j++];
    }

    lhd0 = 0;
    
    for(st u=0; u<vvg.size(); u++) {
        Vg &vg = vvg[u];

        spr[u] = 0;
        for(st i=0; i<vg.size(); i++) {
            vg[i].p[0] = hp.m[vg[i].v];
            vg[i].p[1] = hp.m[vg[i].w];
                // update per subject probability
            spr[u] += vg[i].p[0] * vg[i].p[1];
        }

            // update initial sample log-likelihood
        lhd0 += log(spr[u]);
    }

        // compute initial gametic counts
    CountEgs(eg, hp, spr, vvg, vsvi);
}

void ReCount(Md& eg, Hap& hp, Vd& spr, double& lhd0, 
             Vvg& vvg, const Vsvi& vsvi)
{
    hp.m.clear();
    for(Svi::iterator i=hp.s.begin(); i != hp.s.end(); i++) {
        hp.m[*i] = eg[*i] / (2.0 * vvg.size());
    }

    lhd0 = 0;
    
    for(st u=0; u<vvg.size(); u++) {
        Vg &vg = vvg[u];
        spr[u] = 0;
        for(st i=0; i<vg.size(); i++) {
            vg[i].p[0] = hp.m[vg[i].v];
            vg[i].p[1] = hp.m[vg[i].w];
                // update per subject probability
            spr[u] += vg[i].p[0] * vg[i].p[1];
        }
            // update sample log-likelihood
        lhd0 += log(spr[u]);
    }
        // update gametic counts
    CountEgs(eg, hp, spr, vvg, vsvi);
}

// create a vector of sets of possible haplotypes for each person
void GetIndHaps(const Vvg& vvg, Vsvi& vsvi)
{
    if(vsvi.size() != vvg.size()) vsvi.resize(vvg.size());
    for(st u=0; u<vvg.size(); u++) {
        const Vg &vg = vvg[u];
        Svi s;
        for(st i=0; i<vg.size(); i++) {
            s.insert(vg[i].v);  
            s.insert(vg[i].w);
        }
        vsvi[u] = s;
    }    
}

double ExpMax(Vvg& vvg, Vsvi& vsvi, Hap& hp, Md& eg, Vg& holder,
              Vdu& vdu, double diri_par, int restarts, double tol, int out)
{
    GetGametes(holder, vvg, vdu, diri_par);
    Vd spr(vvg.size()); // per subject probabilities
    GetIndHaps(vvg, vsvi);
    Vd lhoods(restarts);

    Md eg_m;
    double lhdi_m=0;

    for(int i=0; i<restarts; i++) {
        double lhd0, lhdi;

        Initiate(eg, hp, spr, lhd0, vvg, vsvi, diri_par);
        lhdi = lhd0;

        do {
            lhd0 = lhdi;
            ReCount(eg, hp, spr, lhdi, vvg, vsvi);
        } while(fabs(lhd0 - lhdi) > tol && (lhdi > lhd0));        

        if(i==0) {
            lhdi_m = lhdi;
            eg_m = eg;
        }
        else if(lhdi > lhdi_m) {
            eg_m = eg;
            lhdi_m = lhdi;
        }
    }

    eg = eg_m;
    Vd cnts(hp.s.size() + 1);
    cnts[0] = lhdi_m;
    int j=1;
    for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) {
        cnts[j++] = eg[*u];
    }

    cout << "Estimated haplotype frequencies:\n";
    {
        int j=1;
        for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) {
            if(cnts[j] > 0) {
                const Vi& cu = *u;
                for(st i=0; i < cu.size(); i++) {
                    cout << cu[i];
                    if(int(i) < int(cu.size())-1) cout << "_";
                }
                cout << " :\t " << cnts[j] / (2.0*vvg.size()) << "\n";
            }
            ++j;
        }
        cout << "\nMax log-likelihood among " << restarts;
        cout << " restarts: " << cnts[0] << "\n\n";
    }

    if(out)
    {
        cout << "\nProbabilities for each person\n\n";
        for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) 
        {
            if(eg[*u] < IgnoreFreq) continue;
            const Vi& cu = *u;
            for(st i=0; i < cu.size(); i++) {
                cout << cu[i];
                if(int(i) < int(cu.size())-1) cout << "_";
            }
            cout << "; ";
        }
        cout << "\n\n";
    }
    
    return cnts[0];
}

double Betai(double a, double b, double x)
{
    double bt;
    
    if (x < 0.0 || x > 1.0) {
        cerr << "Bad x in routine Betai: " << x << "\n";
        return -999;
    }
    
    if (x == 0.0 || x == 1.0) bt=0.0;
    else
        bt=exp(gammln(a+b)-gammln(a)-gammln(b)+a*log(x)+b*log(1.0-x));
    if (x < (a+1.0)/(a+b+2.0))
        return bt*betacf(a,b,x)/a;
    else
        return 1.0-bt*betacf(b,a,1.0-x)/b;
}

double T_pv(double t, double df)
{
    double x = df / (df + t*t);
    return Betai(df/2.0, 0.5, x);
}

double F_pv(double fstat, double reg_df, double err_df)
{
    double f = err_df / (err_df + reg_df * fstat);
    return Betai(err_df/2.0, reg_df/2.0, f);
}

// compute generalized inverse of X'X through SVD
void Ginverse(Matrix& XXi, Matrix& X, int rank)
{
    Matrix XXt = X.t() * X;
    Matrix Q(XXt.Ncols(), XXt.Ncols());
    Matrix P(XXt.Nrows(), XXt.Ncols());
    DiagonalMatrix S(XXt.Ncols());
    SVD(XXt, S, P, Q);
    for(int i=1; i<=rank; i++) S(i,i) = 1.0 / S(i,i);
    for(int i=rank+1; i <= S.Nrows(); i++) S(i,i) = 0;
    XXi = Q * S * P.t();
}

int FindCollinearities(Vi& rm, Matrix& X)
{
    int r=X.Nrows();
    int c=X.Ncols();
    rm[2]=1;
    
    for(int j=2; j<=c-1; j++) {
        for(int m=j+1; m<=c; m++) {
            int keep1=0;
            for(int i=1; i<=r; i++) {
                if(X(i, j) != X(i, m)) {
                    keep1 = 1;
                    break;
                }
            }
            if(keep1) {
                rm[m] = 1;
            }
            else {
                rm[j] = 0;
                rm[m] = 1;
            }
        }
    }
    int rt=0;
    for(st i=0; i<rm.size(); i++) rt += rm[i];
    return rt;
}

void GetMats(Matrix& rX, Matrix& X, Matrix& X1, Matrix& XX1i,
             vector<Matrix>& X2, vector<Matrix>& XX2i,
             Vdu& vdu, Vvg& vvg, Hap& hp, Md& eg, Vsvi& vsvi)

{
    const int r = vvg.size();
    const int c = hp.s.size();
    X.ReSize(r, c+1);
  
    for(st i=0; i<vvg.size(); i++) {
        X(i+1, 1) = 1;
        int j=2;
        for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) 
        {
            if(vsvi[i].find(*u) != vsvi[i].end()) {
                Vi dst(u->size());
                Reconstruct(dst, *u, vvg[i][0].g);
                double cprob1 = eg[*u];
                double cprob2 = eg[dst];
                X(i+1, j) = cprob1 * cprob2;
            }
            else {
                X(i+1, j) = 0;
            }
            ++j;
        }
            // rescale
        double sum=0;
        for(j=2; j <= X.Ncols(); j++) sum += X(i+1, j);
        for(j=2; j <= X.Ncols(); j++) X(i+1, j) /= sum;
    }

    Svi::iterator u=hp.s.begin();
    for(int j=2; j <= X.Ncols(); j++) {
        const Vi& cu = *u;
        cout << "Haplotype ";
        for(st i=0; i < cu.size(); i++) {
            cout << cu[i];
            if(int(i) < int(cu.size())-1) cout << "_";
        }

        double wsum=0, mean=0;
        for(int i=1; i <= X.Nrows(); i++) {
            mean += vdu[i-1].rsp * X(i, j);
            wsum += X(i, j);
        }
        mean /= wsum;
        
        cout << " mean:\t " << mean << "\n";
        ++u;
    }
    
    Vi rm(X.Ncols() + 1);
    fill(rm.begin(), rm.end(), 0);
    int nc = FindCollinearities(rm, X);
    
    rX.ReSize(r, nc+1);
    for(int i=1; i<=r; i++) {
        int v=1;
        for(int j=1; j<=X.Ncols(); j++) {
            if(j==1) rX(i,j) = X(i,j);
            else if(j>1 && rm[j]) rX(i, ++v) = X(i,j);
        }
    }    

    X1 = rX.SubMatrix(1, rX.Nrows(), 1, rX.Ncols()-1);
    Ginverse(XX1i, X1, X1.Ncols());
    XX1i = XX1i * X1.t();

    XX2i.resize(c);
    X2.resize(c);

    for(int i=1; i<=c; i++) {
        X2[i-1].ReSize(X.Nrows(), 2);
        for(int j=1; j<=X2[i-1].Nrows(); j++) {
            X2[i-1](j,1) = 1;
            X2[i-1](j,2) = X(j, i+1);
        }
        Ginverse(XX2i[i-1], X2[i-1], X2[i-1].Ncols());
        XX2i[i-1] = XX2i[i-1] * X2[i-1].t();
    }
}

double Regress(Matrix& X, Matrix& X1, Matrix& XX1i,
               vector<Matrix>& X2, vector<Matrix>& XX2i,
               double nruns, Vd& stats, Vdu& vdu, Vvg& vvg, Hap& hp, Md& eg,
               Vsvi& vsvi, int out=0)
{
    const int r = vvg.size();
    const int c = hp.s.size();

    ColumnVector Y(r);
    for(st i=0; i<vvg.size(); i++) {
        Y(i+1) = vdu[i].rsp;
        if(out) {
            cout << mt << Y(i+1) << " ";
            for(int j=2; j <= X.Ncols(); j++) cout << mt << X(i+1, j) << " ";
            cout << "\n";
        }
    }
        // param est
    ColumnVector b1 = XX1i * Y;

        // reduced model SS
    double y2 = pow(Y.Sum(), 2);
    double redss = y2 / Y.Nrows();

        // total corrected sum of squares
    Matrix tssc = Y.t()*Y - redss;
    double dtssc = tssc(1, 1);

        // model sum of squares
    Matrix ssm = b1.t() * X1.t() * Y;
    double dssm = ssm(1, 1);

        // regression SS
    double reg_ss = dssm - redss;
    int reg_df = X1.Ncols() - 1;
    int err_df = (Y.Nrows() - 1) - reg_df; // tot df - reg df

        // error SS
    double err_ss = dtssc - reg_ss;
    double reg_ms = reg_ss / reg_df;
    double err_ms = err_ss / err_df;

    stats.resize(b1.Nrows()+1);

        // F stat for the model after fitting the intercept
    double fstat = reg_ms / err_ms;
    double pv_fstat = F_pv(fstat, reg_df, err_df);
    
    stats[0] = nruns > 0 ? fstat : pv_fstat;
    
        // T stats for each coefficient
    for(int i=1; i<=c; i++) {
        ColumnVector b2 = XX2i[i-1] * Y;
        ssm = b2.t() * X2[i-1].t() * Y;
        dssm = ssm(1, 1);
        reg_ss = dssm - redss;
        reg_df = X2[i-1].Ncols() - 1;
        err_df = (Y.Nrows() - 1) - reg_df;
        err_ss = dtssc - reg_ss;
        reg_ms = reg_ss / reg_df;
        err_ms = err_ss / err_df;
        double b_fstat = reg_ms / err_ms;
        double b_pv = F_pv(b_fstat, reg_df, err_df);
        stats[i] = nruns > 0 ? b_fstat : b_pv;
    }
    return pv_fstat;
}

void ZeroCols(Hap& hp, const Vvg& vvg, Vsvi& vsvi, Md& eg)
{
    Hap hp0;
    for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) {
        set<double> res;
        for(st i=0; i<vvg.size(); i++) {
            if(vsvi[i].find(*u) != vsvi[i].end()) {
                Vi dst(u->size());
                Reconstruct(dst, *u, vvg[i][0].g);
                double cprob1 = eg[*u];
                double cprob2 = eg[dst];
                res.insert(cprob1*cprob2);
            }
        }
        if(res.size() > 1) hp0.s.insert(*u);
    }
    hp = hp0;
}

void Weedout(Hap& hp0, Hap& hp1, Md& eg, Vvg& vvg, Vsvi& vsvi)
{
    Hap hp = hp1;
        //ZeroCols(hp, vvg, vsvi, eg);

    for(Svi::iterator u=hp.s.begin(); u != hp.s.end(); u++) 
    {
        if(eg[*u] >= IgnoreFreq) hp0.s.insert(*u);
    }
}

void Shuffle(Vdu& vdu)
{
    int n = vdu.size();
    for(int i=0; i<n; i++) swap(vdu[i].rsp, vdu[i+Ga(n-i)].rsp);
}

int NCols(const char* fname)
{
    ifstream ifs(fname);
    string buf;
    getline(ifs, buf);
    vector<string> w;
    int n = GetNumWords(buf, w, " \t\r");
    return n;
}

int Translate_if_strings(const char *oname, const char *iname, int nlo, string mis)
{
    typedef map <string, int, less<string> > Ms;
    typedef set <string> Ss;
    typedef vector <string> Vs;

    ifstream Cin(iname);

    Ms *ms = new Ms [nlo];
    Ss *ss = new Ss [nlo];
    Vs *vs = new Vs [nlo*2];

    Vs resps;
    int all_nums=1;
    
    while (Cin) {
        string response;
        if(!(Cin >> response)) break;
        resps.push_back(response);
        for (int i=0; i < nlo*2; i++) {
            string s;
            if(!(Cin >> s)) break;
            if(s != mis && Is_An_Integer(s.c_str()) == 0) all_nums=0;
            ss[i/2].insert(s);
            vs[i].push_back(s);
        }
    }

    if(all_nums) return 1;

    ofstream Cout(oname);

    for (int i=0; i<nlo; i++) {
        int v=0;
        for (Ss::iterator j=ss[i].begin(); j != ss[i].end(); j++) {
            if(*j != mis) ms[i][*j] = v++;
            else ms[i][*j] = -1;
        }
    }
    
    for (size_t j=0; j<vs[0].size(); j++) {
        Cout << resps[j] << " ";
        for (int i=0; i<nlo*2; i+=2) {
            int a[2] = { 1+ms[i/2][vs[i][j]], 1+ms[i/2][vs[i+1][j]] };
            if(a[0] == 0) a[0] = -1;
            if(a[1] == 0) a[1] = -1;
            if (a[0] > a[1]) swap(a[0], a[1]);
            Cout << setw(3) << a[0] << " " << setw(3) << a[1] << "  ";
        }
        Cout << endl;
    }
    delete [] ms;
    delete [] ss;
    delete [] vs;

    return 0;
}

int main(int ac, char **av)
{
    if(ac != 7) {
        cerr << "Supply the following arguments in the same order:\n\n";
        cerr << "num_restarts precision missing_value file_name num_runs rnd_seed.\n\n";
        cerr << "num_restarts: 1,2,3... (number of EM restarts).\n\n";
        cerr << "If num_restarts is even, will output per-subject probabilities.\n\n";
        cerr << "precision: .001 or such (logL difference EM stopping condition).\n\n";
        cerr << "missing_value: '?' (quote special characters!) or 0, or whatever\n";
        cerr << "the missing value string is.\n\n";
        cerr << "file_name: space-separated data file. First column is the response,\n";
        cerr << "then markers, 2 columns per marker. If some or all allele names aren't\n";
        cerr << "numbers, they will be re-coded to numbers in the output.\n\n";
        cerr << "num_runs: number of shufflings. If 0, compute asymptotic test.\n\n";
        cerr << "rnd_seed: integer seed for the random number generator\n";
        
        return 1;
    }
    
    int restarts = atoi(av[1]);
    double precision = atof(av[2]);
    string miss = av[3];
    const char *fname = av[4];
    int nruns = atoi(av[5]);
    unsigned long seed = atol(av[6]);

    const double diri_par = 1;

    ios::sync_with_stdio();

    int Out=1;
    if(miss == "sim" || IsOdd(restarts)) Out=0;

    Ga.Init(seed);
    Vdu vdu;

    int n = NCols(fname)-1;
    assert(IsOdd(n) == 0);
    n /= 2;

    char* tfn = tmpnam(0);
    int all_nums = Translate_if_strings(tfn, fname, n, miss);

    if(all_nums) {
        ifstream ifs(fname);
        ReadGnt(vdu, n, miss, ifs);
        ifs.close();
    }
    else {
        ifstream ifs(tfn);
        ReadGnt(vdu, n, "-1", ifs);
        ifs.close();
        remove(tfn);
    }
    
    Vg holder;
    Vd st_0, st_i;
    Vi cnt;
    Hap hp, hp0;
    Md eg;
    Vsvi vsvi;
    Vvg vvg;

    RunIndicator ri(nruns/20);
    ExpMax(vvg, vsvi, hp, eg, holder, vdu, diri_par, restarts, precision, Out);
    Weedout(hp0, hp, eg, vvg, vsvi);

    Matrix rX, X, X1, XX1i;
    vector<Matrix> X2, XX2i;
    GetMats(rX, X, X1, XX1i, X2, XX2i, vdu, vvg, hp0, eg, vsvi);

    double pv_fstat = Regress(X, X1, XX1i, X2, XX2i, nruns, st_0, 
                              vdu, vvg, hp0, eg, vsvi, Out);

    cnt.resize(st_0.size());
    fill(cnt.begin(), cnt.end(), 0);
    
    for(int i=0; i<nruns; i++) {
        Shuffle(vdu);
        Regress(X, X1, XX1i, X2, XX2i, nruns, st_i, vdu, vvg, hp0, eg, vsvi);
        for(st j=0; j<st_0.size(); j++) {
            if(st_i[j] >= st_0[j]) ++cnt[j];
        }
        if(miss != "sim") ri.go();
    }

    if(nruns && miss != "sim") ri.flush();

    string perm = nruns > 0 ? "Permutation" : "Asymptotic";
    double dnruns = nruns, pv = nruns > 0 ? cnt[0]/dnruns : st_0[0];
    cout << perm << "-based tests\n";
    printf("Overall p-value is %g\n", pv);
    st j=1;
    for(Svi::iterator u=hp0.s.begin(); u != hp0.s.end(); u++)
    {
        if(j >= st_0.size()) break;
        pv = nruns > 0 ? cnt[j]/dnruns : st_0[j];
        cout << "Ind haplotype ";
        const Vi& cu = *u;
        for(st i=0; i < cu.size(); i++) {
            cout << cu[i];
            if(int(i) < int(cu.size())-1) cout << "_";
        }
        cout << " :\t ";
        printf("p-value is %g\n", pv);
        ++j;
    }
    
    return pv_fstat < 0;
}
